Require TransitionSystems.
Require TheoryList_compl.
Require PolyList.
Require Peano_dec.
Require Minus.
Require Relations_2.
Require Compare.
Require Compare_dec.
Require Lt.
Require Le.
Require Export TheoryList.
Require Nat_compl.

Definition register := nat.

Inductive instruction : Set :=
   Succ : register -> instruction
 | DecJump : register -> nat -> instruction.

Inductive uses_at_most [r:register] : instruction -> Prop :=
   okSucc : (r':register)(lt r' r) -> (uses_at_most r (Succ r'))
 | okDecJump : (r':register)(n:nat)(lt r' r) -> (uses_at_most r (DecJump r' n)).

Record RAM : Set :=
 { registers_number : nat ;
   program : (list instruction) ;
   program_ok : (AllS (uses_at_most registers_number) program)
 }.

Record configuration : Set :=
 { program_counter : nat ;
   registers : register -> nat
 }.

(* The Operational Semantics *)

Definition update :=
 [registers : register -> nat][r: register][v:nat]
  [r': register]
   Cases (Peano_dec.eq_nat_dec r r') of
      (left _) => v
    | (right _) => (registers r')
   end.

Inductive step' [pc:nat] :
  instruction -> configuration -> configuration -> Prop
 :=
   Succ_step :
    (r:register)(c:configuration)
      [registers := (registers c)]
      [new_c :=
       (Build_configuration (S pc) (update registers r (S (registers r))))]
      (step' pc (Succ r) c new_c)
 | DecJump_step :
    (r:register)(pc':nat)(c:configuration)
      [registers := (registers c)]
      [oldval := (registers r)]
      [registers' := (update registers r (Minus.minus oldval (S O)))]
      [pc'' :=
        Cases oldval of
           O => pc'
         | (S _) => (S pc)
        end]
      [new_c := (Build_configuration pc'' registers')]
       (step' pc (DecJump r pc') c new_c).

Definition step :=
 [r:RAM ; c,c':configuration]
  [pc := (program_counter c)]
  [program := (program r)]
  [dummy := (Succ O)] (* It will be never used. Not very clean. *)
   (le (S pc) (length program)) /\
    [instr := (nth pc program dummy)]
     (step' pc instr c c').

Definition step_plus :=
 [r:RAM]
  (TransitionSystems.step_plus ? (step r)).

Definition step_star :=
 [r:RAM]
  (TransitionSystems.step_star ? (step r)).

Theorem step'_deterministic:
 (pc:nat)(i:instruction)
  (deterministic ? (step' pc i)).
Proof.
 Intros;
 Induction i;
  [ Unfold deterministic;
    Induction x;
    Intros;
    Inversion H;
    Inversion H0;
    Unfold new_c;
    Unfold new_c0;
    Cut
     (update registers1 r0 (S (registers1 r0))) =
     (update registers2 r1 (S (registers2 r1)));
     [ Intros Hcut;
       Rewrite Hcut;
       Reflexivity
     | Rewrite H2;
       Rewrite H5;
       Unfold registers1;
       Unfold registers2;
       Rewrite H1;
       Rewrite H4;
       Reflexivity
     ]
  | Unfold deterministic;
    Induction x;
    Intros;
    Inversion H;
    Inversion H0;
    Unfold new_c;
    Unfold new_c0;
    Cut (registers'=registers'0);
    [ Intro Hequal_registers;
      Rewrite <- Hequal_registers;
      Apply (f_equal2 ? ? ? Build_configuration);
       [ Unfold pc'';
         Unfold pc''0;
         Unfold oldval;
         Unfold oldval0;
         Unfold registers1;
         Unfold registers2;
         Rewrite H2;
         Rewrite H6;
         Rewrite H3;
         Rewrite H7;
         Rewrite H1;
         Rewrite H5;
         Reflexivity
       | Reflexivity]
    | Unfold registers';
      Unfold registers'0;
      Unfold oldval;
      Unfold oldval0;
      Unfold registers1;
      Unfold registers2;
      Rewrite H2;
      Rewrite H6;
      Cut c=c0;
       [ Intro Hequal_c_c0;
         Rewrite Hequal_c_c0;
         Reflexivity
       | Rewrite H1;
         Rewrite H5;
         Reflexivity
       ]
    ]
  ].
Qed.

Theorem step_deterministic:
 (r:RAM)(deterministic ? (step r)).
Proof.
 Intro;
 Unfold deterministic;
 Unfold step;
 Intros;
 Elim H;
 Intros;
 Elim H0;
 Intros;
 Clear H H0 H1 H3;
 Generalize
  (step'_deterministic (program_counter x)
    (nth (program_counter x) (program r) (Succ (0))));
 Intro step'_determ;
 Unfold deterministic in step'_determ;
 Apply (step'_determ x y z);
 Assumption.
Qed.

Theorem step'_total :
 (pc:nat)(i:instruction)
  (total ? (step' pc i)).
Proof.
 Intros;
 Induction i;
  (* Both cases *)
  Unfold total;
  Intro;
  EApply exist;
  Constructor.
Qed.

Theorem step_ok_with_registers:
 (r:RAM; c,c':configuration)
  (step r c c') ->
   (n:register)
    (gt n (registers_number r)) -> (registers c n) = (registers c' n).
Proof.
 Intros;
 Elim H;
 Clear H;
 Intros;
 Cut
  (uses_at_most (registers_number r)
    (nth (program_counter c) (program r) (Succ (0))));
 [ Intro Huses_at_most;
   Inversion Huses_at_most;
    [ Rewrite <- H2 in H1;
      Inversion H1;
      Unfold new_c;
      Simpl;
      Rewrite H5;
      Unfold registers0;
      Rewrite H4;
      Unfold update;
      Elim (eq_nat_dec r' n);
       [ Intro;
         ElimType False;
         Apply (Rocq.THREE_GAP.Nat_compl.lt_not_eq r' n);
          [ Unfold gt in H0;
            Exact (lt_trans ? ? ? H3 H0)
          | Assumption
          ]
       | Reflexivity
       ]
    | Rewrite <- H2 in H1;
      Inversion H1;
      Unfold new_c;
      Simpl;
      Unfold registers';
      Unfold update;
      Unfold registers0;
      Rewrite H4;
      Rewrite H5;
      Elim (eq_nat_dec r' n);
       [ Intro;
         ElimType False;
         Apply (Rocq.THREE_GAP.Nat_compl.lt_not_eq r' n);
          [ Unfold gt in H0;
            Exact (lt_trans ? ? ? H3 H0)
          | Assumption
          ]
       | Reflexivity
       ]
    ]
 | Clear H1 H0 n c';   (* Probably this should be a lemma *)
   Generalize (program_ok r);
   Intro;
   Generalize (nth_In (Succ (0)) H);
   Intro;
   Apply (AllS_In ? ? ? ? H0 H1)
 ].
Qed.

Definition stuck :=
 [r:RAM](TransitionSystems.stuck ? (step r)).

Theorem stuck_iff_ge_pc_program_length:
 (r:RAM)(c:configuration)
  (stuck r c) <-> (ge (program_counter c) (length (program r))).
Proof.
 Split;
  [ Intro;
    Elim (le_lt_dec (length (program r)) (program_counter c));
     [ Intro;
       Exact a
     | Intro;
       Cut ~(stuck r c);
        [ Intro;
          Elim (H0 H)
        | Unfold not;
          Intro;
          Elim H;
          Generalize
           (step'_total
             (program_counter c)
             (nth (program_counter c) (program r) (Succ (0))) c
           ) ;
          Induction 1;
          Intros;
          Exists x;
          Split;
           [ Assumption
           | Assumption
           ]
        ]
     ]
  | Intro;
    Unfold stuck;
    Unfold not;
    Intro;
    Elim H0;
    Clear H0;
    Intros;
    Elim H0;
    Intros;
    Apply (le_not_lt ? ? H H1)
  ].
Qed.

Theorem stuck_decidable:
 (r:RAM)(c:configuration)
  {(stuck r c)}+{~(stuck r c)}.
Proof.
 Intros;
 Elim (stuck_iff_ge_pc_program_length r c);
 Intros;
 Elim (le_lt_dec (length (program r)) (program_counter c));
  [ Left;
    Exact (H0 a)
  | Right;
    Intro;
    Generalize (H H1);
    Intro;
    Apply (le_not_lt ? ? H2 b)
  ].
Defined.

Theorem step_decidable:
 (r:RAM)(c:configuration)
  {(EX c'| (step r c c'))}+{~(EX c' | (step r c c'))}.
Proof.
 Intros;
 Unfold step;
 Elim (le_lt_dec (length (program r)) (program_counter c));
  [ Intro;
    Right;
    Intro;
    Decompose [ex and] H;
    Elim le_not_lt with 1 := a;
    Assumption
  | Left;
    Elim (nth (program_counter c) (program r) (Succ (0)));
     [ Intros;
       EApply ex_intro;
        Split;
	 [ Assumption
	 | Constructor
	 ]
     | Intros;
       EApply ex_intro;
        Split;
	 [ Assumption
	 | Constructor
	 ]
     ]
  ].
Defined.

Definition reduces :=
 [r:RAM](TransitionSystems.reduces ? (step r)).

(* Reduces is deterministic, but not total. *)

Theorem reduces_deterministic:
 (r:RAM)(deterministic ? (reduces r)).
Proof.
 Intro;
 Unfold deterministic;
 Induction 1;
 Clear H;
 Induction 3;
 Clear H1;
 Intros;
  Induction H;
   [ Induction H1;
      [ Reflexivity
      | Elim H0;
        Exists y;
        Trivial
      ]
   | Induction H1;
      [ Elim H2;
        Exists y;
        Trivial
      | Generalize (step_deterministic ? ? ? ? H H1);
        Intro;
        Apply HrecH;
         [ Assumption
         | Rewrite <- H5 in H4;
           Exact H4
         ]
      ]
   ].
Qed.

Definition RAM_configuration_equality :
 nat -> configuration -> configuration -> Prop
:=
 [registers_no:nat ; H,H':configuration]
  Cases H of
   (Build_configuration program_counter registers) => 
     Cases H' of
      (Build_configuration program_counter' registers') => 
        program_counter=program_counter' /\
        (r:register)
         (lt r registers_no) ->
          (registers r)=(registers' r)
     end
  end.
